@@ -14,4 +14,4 @@
 <================"""
 from .Face import Face, DrawingSpec
 from .FaceAnalyzer import FaceAnalyzer
-from .Helpers import buildCameraMatrix, rodriguezToRotationMatrix
\ No newline at end of file
+from .Helpers import buildCameraMatrix, faceOrientation2Euler
\ No newline at end of file
